//  $Id: pfeap.hh,v 1.12 2011-05-20 17:32:59 madams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2003 by Mark F. Adams 
//  Filename:           pfeap.hh
//
#ifndef  __PFEAP_H__
#define  __PFEAP_H__

#include "athena.hh"

#if defined(PROM_HAVE_SILO)
#include <silo.h>
#endif


////////////////////////////////////////////////////////////////////////
// OlyCoord
////////////////////////////////////////////////////////////////////////
class OlyCoord
{
public:
  OlyCoord(){ m_x[0]=0.; m_x[1]=0.; m_x[2]=0.; }
  OlyCoord(double x, double y, double z){ 
    m_x[0]=x; m_x[1]=y; m_x[2]=z; 
  }
  double &operator[](int ind){
    if(ind<0 || ind>2 ) exit(13);
    return m_x[ind];
  }
  double operator[](int ind)const{
    if(ind<0 || ind>2 ) exit(13);
    return m_x[ind];
  }
private:
  double m_x[3];
};

////////////////////////////////////////////////////////////////////////
// OlyTri
////////////////////////////////////////////////////////////////////////
class OlyTri
{
  public:
  OlyTri(){ m_id[0]=0.; m_id[1]=0.; m_id[2]=0.; }
  int &operator[](int ind){
    if(ind<0 || ind>2 ) exit(13);
    return m_id[ind];
  }
  int operator[](int ind)const{
    if(ind<0 || ind>2 ) exit(13);
    return m_id[ind];
  }
private:
  int m_id[3];
};

class pFEAP;
////////////////////////////////////////////////////////////////////////
// class SurfaceMesh
////////////////////////////////////////////////////////////////////////
class SurfaceMesh
{
  friend class pFEAP;
public:
  SurfaceMesh(const PromComm &comm);
  virtual ~SurfaceMesh();
  int Init( char *filename );
  // data access methods
  int getNTri()const {return m_ntri;}
  int getNNodes()const {return m_nNodes;}
  int AddSurfPres( PromVector *res )const;
  //
  const PromComm &Comm_; 
  // mesh and data on mesh
  double *m_x_curr[3]; // [m_nNodes]
  double *m_v_curr[3]; // [m_nNodes]
  double *m_presForce[3];// [m_nNodes]
  int *m_tri[3];       // [m_ntri]
  // initial state of surface mesh
  OlyCoord *m_x_0; // [m_nNodes] original X and V
  OlyCoord *m_v_0;
private:
  int *m_fids;  // [m_nNodes] original (before partitioning) IDs on full mesh
  int *m_gids;  // [m_nNodes] global IDs on full mesh
  int m_ndf;
  int m_ntri;   // number of triangles in surface mesh
  int m_nNodes; // number of nodes in surface mesh
};
 
class Prometheus_LinSysCore;
////////////////////////////////////////////////////////////////////////
//! Interface to FE application
////////////////////////////////////////////////////////////////////////
class FEComponentIterface
{
public:
  virtual int SetUp( SurfaceMesh *smesh, int numParams, char** params , double ) = 0;
  virtual int Advance( const double, const int, SurfaceMesh * ) = 0;
  virtual int Finalize() = 0;
};

////////////////////////////////////////////////////////////////////////
//! Interface to FE application -- FEAP
////////////////////////////////////////////////////////////////////////
class pFEAP : public FEComponentIterface
{
public:
  pFEAP(const PromComm &comm, PromOptions &opts, PromPerfMonitor &perf);
  virtual ~pFEAP();
  // FEDataClient
  int SetUp( SurfaceMesh *smesh, int numParams, char** params, double );
  int Advance( const double dt, const int, SurfaceMesh *smesh = 0 );
  int Finalize();
private:
  int SetUpFEAP();
  int SetUpPrometheus(double);
  // output
  int InitializeDB( const int nsteps, char *prestr = NULL )const;
  int StepDB( int step, const int nsteps, double time, const char *prestr )const;
  int CreateMultiDB( int s,const int nsteps, double t, const char *p)const;
#if defined(PROM_HAVE_SILO)
  int CreateDB( DBfile **dbfile_out, int step,  const int nsteps,
		double time, const char *prestr ) const;
  int PutMesh( DBfile*, int, double )const;
#endif
  int FinalizeDB( int nsteps, double dt, char *prestr=NULL )const;
  //
  int GetMeshSpecs( int *nNds, int *nElems, int *nLocEls, int * )const;
  int GetNElemBlks(int *nsets)const{ *nsets=1; return 0;}
  int GetNMaterials( int *nm )const;
  int GetNElemsDegree(int ElBlkID, int *nconn)const;
  int GetNNodes( int *nn )const;
  int GetElemMeshBlk( int ElBlkID, double coords[], 
		      int matids[], int elems[] )const;
  int GetElemBlkAtr(int ElBlkID, int nel, float * )const;
  int GetNElemAtr(int ElBlkID, int *na )const;
  int GetNodeDisp( float array[], int ) const;
  int GetSurfXV( SurfaceMesh *smesh )const;
  int SetUpSurface( SurfaceMesh *smesh );
  // FEComponentIterface
  int GetSolution( PromVector *const XX );
  int GetResidual( PromVector *const RR );
  int ComputeResidual();
  int Reac( int nnodes, double );
  int Disp( int nnodes, double );
  int Stress( int nnodes );
  int FormTangent(Prometheus_LinSysCore*const);
  int FormTangentAndResidual(Prometheus_LinSysCore*const);
  //
  //int CopyVectorOut( PromVector *const BB, double const*const array );
  //int CopyVectorIn( const PromVector *const XX, double *const array );
  // methods
  int ReadMesh( const char *feapname, int nlocalelems = 0 );
  int SetUpLinSys( const int npLoc, const int npGst, 
		   const int ghost_gnode[], int &my0Nd );
  int SetDT( double );
  int SetDynamic();
  int StepTime();
 
  // feap displacemetn output
  int WriteDispFile() const;
  int WriteDisp( const int step, int )const;
  // util
  int SetMaps( const int nlocalnd, const int pnode_fnode[],
	       const int ndf, const int my0Nd );
private:
  int Newton( const int, SurfaceMesh *surfacemesh = 0 );
  int GetSolution_private( const PromVector *const XX );
  // data
  AthenaFeap *athena_;
protected:
  IS    lpetsc_gfeapIS_;  // used to put petsc.x back to real feap
  IS    lfeap_gpetscIS_;  // used to put petsc.b out of feap
  IS    afeap_gpetscIS_;  // used to put petsc.x into feap
public:
  int   nfixed_;          // includes ghost nodes
  int   nLocalFixed_;     // includes non-ghost nodes only
  int   nNodes_;
  int   nneq_;            // total equations locally
  int   feapEQIDsz_;      // feapEQIDsz_ = nneq - nLocalFixed_
  int   nLocElems_;
  int  *promeq_appeq_; //[feapEQIDsz_]: X = promeq_appeq_[promEq] + nfixed_
  int  *feapEQID_lid_; //[feapEQIDsz_+1] lid(X) = feapEQID_lid_[X]
  int  *lid_gid_;
  // data
  const PromComm &Comm_; 
  PromOptions &options_;
  PromPerfMonitor &perf_mon_;
protected:
  Prometheus_LinSysCore *linSys_;
  Vec  work_nfl_;
  Vec  work_neq_; 
  IS   identID_nfl_;
  IS   identID_neq_; 
  VecScatter scatter_nfl_;
  VecScatter scatter_neq_;
  // surface mesh interface
  int m_my0;
  int m_ndf;
  int m_nLocNodes;
  int *m_fnode;       // temp cache of original full mesh node IDs
  //
  int  m_nSurfLoc;    // number of surface nodes in my local part. (=0)
  int *m_SurfLocIDs;  // [m_nSurfLoc] local IDs of local part. on surface 
  int *m_globalSurfIDs;// [m_nSurfLoc] global surface IDs of these locals
};

////////////////////////////////////////////////////////////////////////
//! pFERigid
////////////////////////////////////////////////////////////////////////
class pFERigid : public FEComponentIterface
{
public:
  pFERigid(const PromComm &comm, PromOptions &opts, PromPerfMonitor &perf);
  ~pFERigid();
  // FEDataClient
  int SetUp( SurfaceMesh *smesh, int numParams, char** params ,double);
  int Advance( const double, const int, SurfaceMesh * = 0);
  int Finalize();
protected:
  int GetSurfXV( SurfaceMesh *smesh )const;
  // data
  double m_time;
public:
  // data
  const PromComm &Comm_; 
  PromOptions &options_;
  PromPerfMonitor &perf_mon_;
};

////////////////////////////////////////////////////////////////////////
//! pNoFE
////////////////////////////////////////////////////////////////////////
class pNoFE : public FEComponentIterface
{
public:
  pNoFE(const PromComm &comm, PromOptions &opts, PromPerfMonitor &perf){}
  ~pNoFE(){}
  // FEDataClient
  int SetUp( SurfaceMesh *smesh, int numParams, char** params ,double){return 0;}
  int Advance( const double, const int, SurfaceMesh * = 0){return 0;}
  int Finalize(){return 0;}
};
#endif //  __PFEAP_H__
