//  $Id: prom_base.hh,v 1.4 2011-03-14 14:01:21 madams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2000 by Mark F. Adams 
//  Filename:           prom_base.hh
//
#ifndef __PROM_BASE_H__
#define __PROM_BASE_H__

typedef enum{PROMMGMULTIPLICATIVE=0,PROMMGFULL,PROMMGADDITIVE}PromMGAlgo;
typedef enum{PROMKSPGMRES,PROMKSPRICHARDSON,PROMKSPBCGS,PROMKSPCR,PROMKSPTFQMR,PROMKSPPREONLY,PROMKSPCG,PROMKSPCHEB,PROMKSPBREZ,PROMKSPFGMRES,PROMKSPSYMMLQ,PROMKSPMINRES,PROMNUMKSP}PromKSPType;
typedef enum{PROMPCILU,PROMPCLU,PROMPCJACOBI,PROMPCASM,PROMPCNONE,PROMPCMG,PROMPCGAUSSSEIDEL,PROMPCSYMGAUSSSEIDEL,PROMPCNODALASM,PROMNUMPC}PromPCType;

//! PromComm_base
class PromComm_base 
{
public:
  virtual MPI_Comm operator()() const = 0;
  virtual MPI_Comm MPIComm() const = 0;
  virtual int mype() const = 0;
  virtual int np() const = 0;
  virtual int getNewTag() const = 0;
  virtual int getNewTag(MPI_Comm comm) const = 0;
};

//! PromMap_base
#undef __FUNCT__
#define __FUNCT__ "PromMap_base" 
class PromMap_base 
{ 
  friend class PromMatrix;
protected:
  PromMap_base( int nloc, int nloceq, const PromComm_base &comm );
public:
  int ResetMap( int nloc, int nloceq ) {
    int ii;  
    MPI_Comm mpi = comm_(); assert(mpi!=MPI_COMM_NULL); int np = comm_.np();
    // proc_gnode_
    MPI_Allgather( &nloc, 1, MPI_INT, proc_gnode_+1, 1, MPI_INT, mpi );
    for( ii = proc_gnode_[0] = 0 ; ii < np ; ii++ )
      proc_gnode_[ii+1] = proc_gnode_[ii+1] + proc_gnode_[ii];
    // proc_geq_
    MPI_Allgather( &nloceq, 1, MPI_INT, proc_geq_+1, 1, MPI_INT, mpi );
    for( ii = proc_geq_[0] = 0 ; ii < np ; ii++ )
      proc_geq_[ii+1] = proc_geq_[ii+1] + proc_geq_[ii];
    return 0;
  }
  virtual ~PromMap_base(){
    PetscFree( proc_gnode_ ); PetscFree( proc_geq_ );  
    proc_gnode_ = proc_geq_ = NULL;
  }
  virtual int nlocal() const = 0;
  virtual int nglobal() const = 0;
  virtual int nlocalEq() const = 0;
  virtual int nglobalEq() const = 0;
  virtual bool isConstNDF() const = 0;
  virtual int constNDFSize() const = 0;
  virtual int locID_ndf(const int lid) const = 0;
  bool operator==(const PromMap_base &map)const{ 
    assert(this!=NULL);
    return (nlocal()==map.nlocal() && nglobal()==map.nglobal() &&
	    nlocalEq()==map.nlocalEq() && nglobalEq()==map.nglobalEq() ); 
  }
  bool operator!=(const PromMap_base &map)const{ 
    assert(this!=NULL);
    return (nlocal()!=map.nlocal() || nglobal()!=map.nglobal() ||
	    nlocalEq()!=map.nlocalEq() || nglobalEq()!=map.nglobalEq() ); 
  }

  int my0Eq()const{ return proc_geq_[comm_.mype()]; }
  int my0Nd()const{ return proc_gnode_[comm_.mype()]; }
  const int * proc_gnode() const { return proc_gnode_; }
  const int * proc_geq() const { return proc_geq_; }
protected:
  // data
  int *proc_gnode_;
  int *proc_geq_;
public:
  const PromComm_base &comm_;
};

class PromMap;
class PromCRVector;
class PromSolver;
//! class PromVector_base 
/*!
  - base class, pure virtual, interface spec. (PETSc Vec/Petra_vec)
*/
class PromVector_base
{
  friend class PromSolver;
  friend class PromCRVector;
  friend class PromVector;
  friend class PromProjKKTSolver;
public:
  PromVector_base(const PromMap &RowMap) : map_(RowMap)/*,malled_(TRUE)*/ {}
  virtual ~PromVector_base(){}
  // methods
  virtual int GetArray( double **arr ) const = 0;
  virtual int RestoreArray( double **arr ) const = 0;
  virtual int Set( double val ) = 0;
  virtual int Scale( double val ) = 0;
  virtual int Sqrt() = 0;
  virtual int Dot(const PromVector_base*const b, double*const dot)const=0;
  virtual int SetValues( const int nrow, int *rowp, const double *vals,
			 const int add_type) = 0;
  virtual int Norm2( double * const norm ) const = 0;
  virtual int Normalize( double * const norm ) = 0;
  virtual int NormMax( double * const norm ) const = 0;
  virtual int Copy( const PromVector_base * const pvec ) = 0;
  virtual int AXPBY( double a, double b, const PromVector_base *const X)=0;
  virtual int AYPX( double a, const PromVector_base *const X ) = 0;
  virtual int AXPY( double a, const PromVector_base *const X ) = 0;
  virtual int WAXPY(double alpha, const PromVector_base * const X, 
		    const PromVector_base * const Y) = 0;
  virtual int MAXPY(const int nv,double a[],const PromVector_base *const*X)=0;
  virtual int MDot(const int nv,const PromVector_base *const*Xarr,double arr[])const=0;
  virtual int Assembly() = 0;
  virtual int PointwiseMult( const PromVector_base *const X, 
			     const PromVector_base *const Y ) = 0;
  virtual int PointwiseDiv( const PromVector_base *const X, 
			    const PromVector_base *const Y )  = 0;
  virtual int Reciprocal() = 0;
  virtual int getN() const = 0;
  virtual int getn() const = 0;
  virtual int my0() const = 0;
  virtual double operator[](int i) const = 0;
  virtual bool isComplex()const{return FALSE;}
  virtual int Print(char *str=NULL, FILE *file=stderr, int limit=100000)const{
    double *arr; int n = getn();         assert( !isComplex() );
    GetArray( &arr );
    PromContext::printArr( file, arr, (limit > n) ? n : limit,
			   (str==NULL) ? "vector[%d]" : str );
    RestoreArray( &arr );
    return 0;
  }
  //protected:
  const PromMap &map_;
};

class PromPC_base;
//! class PromMatrix_base
/*!
  - base class, pure virtual, interface spec. (PETSc Mat/Petra_Mat)
*/
class PromMatrix_base
{
  friend class PromMatrix;
protected:
  PromMatrix_base( const PromMap &map );
public:
  virtual int Transpose( const PromMap &map, PromMatrix_base ** ) const = 0;
  virtual ~PromMatrix_base(){ DeleteProctable(); }
  //! methods
  virtual int GetLocalNodeRow( const int brow, const int len, int &ncols,
			       int adjacs[], double **Aij = NULL ) const = 0;
  virtual int GetDiagonal( PromVector_base *work ) const = 0;
  virtual int MultTranspose( const PromVector_base * const x,
			     PromVector_base * const b ) const = 0;
  virtual int MultTransposeAdd( const PromVector_base * const x,
				const PromVector_base * const y, 
				PromVector_base * const b ) const = 0;
  virtual int Mult( const PromVector_base * const x, 
		    PromVector_base * const b ) const = 0;
  virtual int KSPMult( const PromVector_base * const x, 
		       PromVector_base * const b ) const = 0;
  virtual int MultAdd( const PromVector_base * const x,
		       const PromVector_base * const y, 
		       PromVector_base * const b ) const = 0;
  virtual int GetRow(int row, int *ncols,int **colpp=NULL,double **arrdat=NULL)const = 0;
  virtual int RestoreRow(int row,int*ncols,int**colpp=NULL,double**arrdat=NULL)
    const = 0;
  virtual int SetValues( const int nrow, int *pgeqi, const int ncol,
			 int *colp, double *vals, const int add_type ) = 0;
  virtual int SetValuesBlocked( const int nrow, int *pgidi, const int ncol,
				int *pcol, double *vs, const int add_type) = 0;
  virtual int ZeroEntries() = 0;
  virtual int ZeroRows( PromIS is, double diags = 0.0 ) = 0;
  virtual int Assembly() = 0;  
  virtual int FlushAssembly() = 0;
  virtual int Shift( double val ) = 0;
  virtual int Scale( double val ) = 0;
  virtual int Scale( const PromVector_base * const x, 
		     const PromVector_base * const b ) = 0;
  virtual int NormFrob( double *const norm ) const = 0;
  virtual int NormInf( double *const norm ) const = 0;
  virtual int AXPY( double val, const PromMatrix_base * const AA ) = 0;
  virtual int SetOption( int opt )  = 0;
  virtual int getN() const = 0;
  virtual int getn() const = 0;
  virtual int getM() const = 0;
  virtual int getm() const = 0;
  virtual int getMNodes() const = 0;
  virtual int getLocalNNZ() const = 0;
  virtual int getGlobalNNZ() const = 0;
  virtual MPI_Comm MPIComm() const = 0;
  virtual int GetGArray( int **arr, int *sz ) const = 0;
  virtual int RestoreGArray( int **arr ) const = 0;
  virtual bool isOK() const = 0;
  virtual bool isAssembled() const = 0;
  virtual bool isSymmetricValued() const = 0;
  virtual bool isSymmetricMatrix() const = 0;
  virtual int Print(char *str=NULL, FILE *file=stderr)const=0;
  virtual int ConvertToSymm()=0;
  virtual int ConvertFromSymm()=0;
  //! non (pure) virtual methods 
  virtual int ComputeLambda(const PromOptions &opt, PromPerfMonitor &perf, 
			    int maxits, PromPC_base *const pc = NULL) 
    const;
  virtual bool isComplex()const{return FALSE;}
  double getDRho() const { assert(this!=NULL); return DRhoA_; }
  int    SetDRho(double r)const{assert(this!=NULL); DRhoA_ = r; return 0;}
  bool   isDirty()const{ return dirty_data_;}
  int    SetDirty(bool b = TRUE) const { dirty_data_ = b; return 0; }
  bool   isIndefinite()const { return (indefinite_ == 1); }
  bool   isDefinite()const { return (indefinite_ == -1); }
  int    SetIndefinite(bool b = TRUE) const { indefinite_ = (b ? 1 : -1); return 0; }
  int    SetIndefinite(int i) const { indefinite_ = i; return 0; } // hard set
  int    DeleteProctable();
  // data
  //protected:
  const PromMap &map_;
  //public:
  //! RAP helpers
  int           numRecv_;
  bool          pRecvDone_;
  int          *recvSzs_;
  int          *recvParts_;
  PromTable<PromTable<PromTable<double*>*>*>  proci_table_gidi_tablej_;
  PromTable<PromTable<PromTable<double*>*>*>  proci_table_gidi_tablej_i_;
  PromTable<int> *gid_ndf_;
protected:
  mutable float DRhoA_;      // Chebyshev parameter
  mutable bool  dirty_data_;
  mutable short int  indefinite_;
};
class PromSeqVector;
//! class PromScatter_base
/*!
  - base class, pure virtual, interface spec.
*/
class PromScatter_base
{
protected:
  PromScatter_base( const PromComm_base &comm ) : comm_(comm) {}
public:
  virtual int Scatter( const PromSeqVector *X, PromVector_base *Y)=0; 
  virtual int Gather( PromSeqVector *X, const PromVector_base *Y)=0; 
  virtual int Gather_r( PromSeqVector *X, const PromVector_base *Y)=0; 
  virtual int Gather_i( PromSeqVector *X, const PromVector_base *Y)=0; 
  virtual int getYIDs( const int **ids )const=0; 
  virtual int restoreYIDs( const int **ids )const=0; 
  virtual int getYID_size()const=0; 
protected:
  const PromComm_base &comm_;
};

/************************************************************************
 *  KKT PCs
 ************************************************************************/
//! class PromPCKKT
/*!
  - Schwarz preconditioner for constrained systems
*/
class PromMatrix;
class PromCRMatrix;
class PromPC;

class PromPCKKT
{
public:
  PromPCKKT(const PromMatrix *CC) : C_(CC), dpc_(NULL), work_(NULL), r_(NULL){}
  virtual ~PromPCKKT();
  virtual int Create( const PromOptions &opts, PromPerfMonitor &perf,
		      const PromMatrix *const AA,
		      const PromCRVector *const XX, bool ilu = FALSE ) = 0;
  virtual int SetOperator( const PromCRMatrix *const KK ) = 0;
  virtual int SetNewOperator( const PromCRMatrix *const KK ) = 0;
  virtual int SetNewPVec( const PromMap &map ) = 0;
  virtual int SetUp();
  virtual int Apply(PromPC_base *const pc, 
		    const PromCRMatrix *const AA,
		    const PromCRVector * const bb, PromCRVector * const xx,
		    const bool zeroguess=TRUE ) = 0;
  virtual const char * getPCString() const = 0;
  virtual int getNumBlocks() const = 0;
  // data
  const PromMatrix *C_;
  PromPC *dpc_; 
protected:
  PromCRVector *work_;
  PromCRVector *r_;
};

/************************************************************************
 *  PromBitArr
 ************************************************************************/

#undef __FUNCT__
#define __FUNCT__ "PromBitArr"
//! class PromPC_base
/*!
  - base PC object 
*/
class PromBitArr 
{
public:
  PromBitArr(int sz) : isz_(sz), count_(0) {
    assert(sizeof(unsigned int) == 4);
    int bsz = sz/32+1;
    PetscMalloc( bsz*sizeof(unsigned int), &data_ ); 
    PetscMemzero( (char*)data_, bsz*sizeof(unsigned int) );
  }
  ~PromBitArr(){ PetscFree( data_ ); }
  bool operator[](const int ii) const { 
    assert(this!=NULL);
    unsigned int a = data_[ii/32], k = ii%32, p = 32 - k - 1;
    unsigned int l = a << p; 
    return ((l>>31) == 1);
  }
  int SetAt(const int ii, bool val){
    assert(this!=NULL);
    unsigned int a = data_[ii/32], k = ii%32, p = 32 - k - 1;
    unsigned int l = a << p; 
    bool b = ((l>>31) == 1);
    if( b != val ){
      unsigned int jj = 1; jj = jj << k;
      if( b ){ a -= jj; count_--; }
      else{ a += jj; count_++; }
      data_[ii/32] = a;
    }
    return 0;
  }
  int SetZero(){
    assert(this!=NULL);int bsz = isz_/32+1;
    PetscMemzero( (char*)data_, bsz*sizeof(unsigned int) );
    count_ = 0;
    return 0;
  }
  int getSize()const{return isz_;}
  int getCount()const{return count_;}
private:
  //! data
  unsigned int *data_;
  unsigned const int isz_;
  unsigned int count_;
};

/*!
  - Fully coupled preconditioner for constrained systems
*/
class PromPCKKT_ASM  : public PromPCKKT
{
public:
  PromPCKKT_ASM( const PromMatrix *const CC, 
		 const int naggs, PromIS *CRIS );
  virtual ~PromPCKKT_ASM();
  virtual int Create( const PromOptions &opts,
		      PromPerfMonitor &perf,
		      const PromMatrix *const AA,
		      const PromCRVector *const XX, bool ilu = FALSE );
  virtual int SetOperator( const PromCRMatrix *const KK );
  virtual int SetNewOperator( const PromCRMatrix *const KK );
  virtual int SetNewPVec( const PromMap &map );
  virtual int Apply(PromPC_base *const pc, 
		    const PromCRMatrix *const AA,
		    const PromCRVector * const bb, PromCRVector * const xx,
		    const bool zeroguess = TRUE );
  virtual const char*getPCString()const{assert(this!=NULL);return "KKT-ASM";}
  virtual int getNumBlocks()const;
protected:
  virtual int CreatePrimAggs( const PromMatrix *const AA, bool ilu );
  int Apply_private( const PromCRVector*const bb, PromCRVector*const xx )const;
#if defined(PROM_USE_PETSC)
  void *aux_mats_;
  void *d_scatters_;
#else // petra
#endif
  // data
  PromIS *PrimIS_;
  PromIS *CRIS_;
  int    *mask_;
  double *zeros_;
  const short int nBlocks_;
  short int nMask_;
public:
  PromBitArr   *lid_isInKKT_; // nodes to skip in primal solver
};

/*!
  - Fully coupled preconditioner for constrained systems
*/ 
class PromPCKKT_Shell  : public PromPCKKT_ASM
{
public:
  PromPCKKT_Shell( const PromMatrix* const AA, const PromMatrix* const CC ) :
    PromPCKKT_ASM( CC, 1, make_is_w_all( CC ) ){}
  virtual int Apply(PromPC_base *const, const PromCRMatrix *const AA,
		    const PromCRVector * const bb, PromCRVector * const xx,
		    const bool zeroguess = TRUE );
  virtual const char * getPCString() const;
  virtual int getNumBlocks() const { return 1; }
protected:
  virtual int CreatePrimAggs( const PromMatrix *const AA, bool ilu );
private:
  static PromIS *make_is_w_all( const PromMatrix* const mat );
};

//! class PromPCKKT_Seg
/*!
  - Dual (projection) preconditioner for constrained systems
*/
class PromPCKKT_Seg : public PromPCKKT
{
public:
  PromPCKKT_Seg( const PromMatrix *const CC, const PromMatrix *const CCt ) : 
    PromPCKKT(CC), CDinvCt_(CCt) {}
  virtual int Create( const PromOptions &opts, 
		      PromPerfMonitor &perf,
		      const PromMatrix *const AA,
		      const PromCRVector *const XX, bool ilu = FALSE );
  virtual int SetOperator( const PromCRMatrix *const KK );
  virtual int SetNewOperator( const PromCRMatrix *const KK );
  virtual int SetNewPVec( const PromMap &map ){ return 0; }
  virtual int Apply( PromPC_base*const, const PromCRMatrix *const AA,
		     const PromCRVector * const bb, PromCRVector * const xx,
		     const bool zeroguess = TRUE );
  virtual const char * getPCString() const { return "KKT-Segregated"; }
  virtual int getNumBlocks() const { return 0; }
  // data
protected:
  //const PromMatrix *A_;
  const PromMatrix *CDinvCt_;
};

/************************************************************************
 *  PromPC_base
 ************************************************************************/

class PromCRMatrix;
class PromGrid;
#undef __FUNCT__
#define __FUNCT__ "PromPC_base"
//! class PromPC_base
/*!
  - base PC object 
*/
class PromPC_base 
{
  friend class Prometheus_LinSysCore;
  friend class PromMG;
  friend class PromPCKKT_ASM;
  friend class Prometheus;
protected:
  PromPC_base(const PromOptions &opt, PromPerfMonitor &perf):
    options_(opt),perf_mon_(perf),KKT_(NULL),numits_(1),
    setup_called_(0),kkt_(NULL){}
public:
  virtual ~PromPC_base();
  // methods
  int SetUp();
  virtual int SetUp_private() = 0;
  //! Apply
  virtual int Apply( const PromVector_base * const bb,
		     PromVector_base * const xx, const bool zerox = TRUE );
  virtual int Apply_private( const PromVector_base*const bb,
			     PromVector_base * const xx, 
			     const bool zerox = TRUE) = 0;
  //! SetOperator methods, set flag for dirty data
  int SetOperator( const PromMatrix_base * const A);
  int SetNewOperator( const PromMatrix_base * const A);
  virtual int SetOperator_private( const PromMatrix_base * const A ) = 0;
  int CreateKKTSmoother(const PromMatrix *const AA, const PromMatrix *const CC,
			const PromCRVector *const XX, const int naggs, PromIS *CRIS,
			const PromMatrix *const CCt ); // 
  int CreateKKTSolver( const PromMatrix *const AA,const PromMatrix *const CC,
		       const PromCRVector *const XX );  // 
  //! utility methods
  virtual const char * getPCString() const = 0;
  virtual const char * getSubPCString() const = 0;
  virtual int getNumBlocks()const{return (kkt_==NULL)?0:kkt_->getNumBlocks();}
  virtual int PrintLevelInfo() const { return 0; } 
  virtual PromPCType getType() const = 0;
  virtual int takeNonZeroGuess()const{ return 0; }
protected:
  virtual int SetNewOperator_private( const PromMatrix_base * const A ) = 0;
public:
  bool operator==( const PromPC_base* a ) const { 
    assert( this != NULL );
    return (a->getType() == getType());
  }
  bool operator!=( const PromPC_base* a ) const { 
    assert( this != NULL );
    return (a->getType() != getType());
  }
  //!
  //int SetNewPVec( const PromMap &map );
  virtual int CreateASM( const int nblocks, PromIS *blockISs ) { 
    PetscPrintf(MPI_COMM_SELF,"[?]%s dummy method ??????\n",
		"PromPC_base::CreateASM");
    return 1;
  }
  virtual int SetType( PromPCType type, const PromVector_base * const v ){
    PetscPrintf(MPI_COMM_SELF,"[?]%s dummy method ??????\n",
		"PromPC_base::SetType");
    return 1;
  }
  int SetNumIts(int nn) { numits_ = nn; return 0; }
  int getNumIts() const { return numits_; }
  // data
  static const char * const PromPCStrings[PROMNUMPC];
protected:
  const PromOptions &options_;
  PromPerfMonitor &perf_mon_;
  const PromCRMatrix       *KKT_;
  short int                 numits_;
  short int                 setup_called_;
public:
  PromPCKKT                *kkt_; // -- depricated
}; // PromPC_base

#endif  // __PROM_BASE_H__
