//  Copyright (c) 2000 by Mark F. Adams 
//  $Id: coar_repart.C,v 1.3 2011-01-22 20:13:39 madams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           coar_repart.C
//
#include "prom_grid.hh"

#undef __FUNCT__  
#define __FUNCT__ "Repartition"
/* PromGrid::Repartition ******************************************************
 *
 *  - coarse grid repartioning methods
 *
 *  INPUT:
 *    -lastgrid: could be NULL
 *   
 *  SIDE EFFECTS:
 *
 *  RETURN: 
 *     - PETSc error code
 */
#if !defined(PROM_HAVE_METIS)
int PromGrid::Repartition( PromGrid * )
{
  return 0;
}
#else 
int PromGrid::Repartition( PromGrid *lgrid )
{
  int ierr;  assert(lgrid!=NULL);
  MPI_Comm wcomm = comm_();

  // repartition
  if( comm_.np() > factor_ && nglobal() > 1 ) { 
    int minloc, nl = isActive() ? nlocal() : 10;
    MPI_Allreduce( &nl, &minloc, 1, MPI_INT, MPI_MIN, wcomm );
    // parMETIS hangs without vertices on a proc....
    if( minloc > 2 ) { 
      perf_mon_.EventBegin( PromContext::prom_logs_[LOG_REPART] ); 
      // repartition
      ierr = Repartition_private( lgrid );   CHKERRQ(ierr);
      perf_mon_.EventEnd( PromContext::prom_logs_[LOG_REPART] ); 
    }
    else if( options_.verbose_>1 && comm_.mype() == 0 ) {
      PetscPrintf( PETSC_COMM_SELF, "[%d]%s(%d) no repartitioning\n",
		   comm_.mype(),__FUNCT__,getLevel());
    }
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "Repartition_private"
/* PromGrid::Repartition_private ******************************
 *
 *  - repartition this (top) grid and reset last parents
 *
 *  INPUT: 
 *    - lgrid:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::Repartition_private( PromGrid *lgrid )
{ 
  const int   mype=comm_.mype(),np=comm_.np();
  int ierr,xx,nloceq,nloc,ii,state,*newproc_gnode=NULL;
  int *lid_part = NULL; 
  int *oldproc_gnode = proc_gnode_;
  int nCommprocs = np/factor_, *part_c = NULL;
  PromTable<int> oldgid_newgid( 3*nNodes_ + 17 );
  PromTable<int> ghostoldgid_newlid( 2*nNodes_ + 17 ); // new gids
  PromList<int> *newLocAdj_oldgid=NULL, *newGhsAdj_oldgid=NULL;
  MPI_Comm wcomm = comm_();
  
  if( isActive() ) {  
    MPI_Comm_size( gComm_, &ii );              assert( ii == nCommprocs ); 
    // partition
    ierr = Repartition( lid_part, part_c, newproc_gnode );  
    if( ierr == 5 ) {
      state = 0; // isolated body!!!
      PetscPrintf(wcomm,"[%d]%s Warning isolated body, abort repartition\n",
		  comm_.mype(),__FUNCT__);
    }
    else {
      CHKERRQ(ierr);
      state = 1;
      // allocate and move most grid data
      ierr = MoveGrid( oldgid_newgid, ghostoldgid_newlid, newproc_gnode,
		       lid_part, part_c, newLocAdj_oldgid, newGhsAdj_oldgid );
      CHKERRQ(ierr);
      proc_gnode_ = newproc_gnode; // new pg
      ierr = PetscFree(part_c); CHKERRQ(ierr); part_c = NULL;
      
      // set new gids on locals, count eqs
      for(xx=nloceq=0, ii = proc_gnode_[mype] ; xx<nLocalNd_ ; xx++, ii++ ){
	nodes_[xx]->index_.setGlobal(ii); 
	nloceq += nodes_[xx]->getNDF();
      } 
      nloc = nLocalNd_;
    }
  }
  else{ state = 1; nloc = nloceq = 0;} 
  if( lgrid->factor_ != factor_ ){
    MPI_Allreduce( &state, &ii, 1, MPI_INT, MPI_MIN, wcomm );
    state = ii;
  }
  if( state == 0 ) return 0; // abort repartitioning
    
  // coolective and global!
  ierr = ResetMap(nloc,nloceq);CHKERRQ(ierr);// collective, deletes petra maps
  assert(stiffness_ == NULL && prolongator_ == NULL);
  
  /*xx = proc_geq_[np];
    MPI_Allgather( &nloceq,1, MPI_INT, proc_geq_+1, 1, MPI_INT, wcomm );
    for(ii=proc_geq_[0]=0;ii<nCommprocs ; ii++)proc_geq_[ii+1]+=proc_geq_[ii];
    assert(xx==proc_geq_[np]);
    if( factor_ != 1 ) { // expand to fill full
    int rbuf[PROM_BSZ];
    for( ii = 0 ; ii < nCommprocs+1 ; ii++ ) rbuf[ii] = proc_geq_[ii];
    for( int pp = 1, tt = 1 ; pp < np ; tt++ ){
    for( ii = 0 ; ii < factor_ ; ii++, pp++ ) proc_geq_[pp] = rbuf[tt];
    }
    }*/ 
  if( isActive() ) {  
    // set node eqs
    for( xx = 0, ii = proc_geq_[mype] ; xx < nLocalNd_ ; xx++ ){
      PromNode *curr = nodes_[xx];   
      curr->setGEQ( ii );
      ii += curr->getNDF();
    }
    // get rest (ghost) data
    ierr = CollectGhostData( oldgid_newgid, ghostoldgid_newlid, 
			     oldproc_gnode, lid_part,
			     newLocAdj_oldgid, newGhsAdj_oldgid );
    CHKERRQ(ierr);
    
    // set global adjacs, make 'ghosts_' (have new gids)
    ierr = GlobalizeAdjac( newproc_gnode[mype], 
			   newLocAdj_oldgid, newGhsAdj_oldgid, oldgid_newgid,
			   ghostoldgid_newlid ); 
    CHKERRQ(ierr);
    //
    delete [] newLocAdj_oldgid;   
    delete [] newGhsAdj_oldgid;   
  }

  if( lgrid->isActive() ) { 
    if( lgrid->factor_ != factor_ ){
      // send grid to slaves
      ierr = DistributeGrid( lgrid, oldgid_newgid ); CHKERRQ(ierr);
    }
    
    // fix last grids parents, send update 
    ierr = FixLastGrid( oldgid_newgid, oldproc_gnode, lid_part, lgrid );
    CHKERRQ(ierr);
    // clean up
    assert( ghosts_.getCount() == nNodes_ - nLocalNd_ );
    if( isActive() ) {
      assert(oldproc_gnode != proc_gnode_);
      PetscFree( oldproc_gnode ); 
      PetscFree( lid_part );
      ierr = GetGhostData( FALSE );                     CHKERRQ(ierr);
    }
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "DistributeGrid"
/* PromGrid::DistributeGrid ******************************
 *
 *  - master send (repartitioned) grid to slaves
 *
 *  INPUT:
 *    - oldgid_newgid:  (out)
 *
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::DistributeGrid( PromGrid *prevG, 
			      PromTable<int> &oldgid_newgid )
{
  const int mype=comm_.mype(),np=comm_.np(),tag=prevG->getNewTag();
  int   ierr,gid,oldgid,xx,tt,ii,proc;
  const int lastfactor=prevG->factor_, masterproc=(mype/factor_)*factor_;
  const int prevpart = mype/lastfactor - 1, nxpart = mype/lastfactor + 1;
  const int endpart = (masterproc + factor_)/lastfactor;  
  TablePos tipos;
  MPI_Status status; MPI_Request sreq;
  int __sbuf[PROM_BSZ], __rbuf[PROM_BSZ]; 
  int *sbuf = ((size_t)__sbuf%sizeof(double)==0) ? __sbuf: __sbuf+1;
  int *rbuf = ((size_t)__rbuf%sizeof(double)==0) ? __rbuf: __rbuf+1; 

  if( isActive() ){  // send updates
    const int sendsz = 2; // < oldgid, new parent >
    // send(tag) proc_gnode
    assert(np+1 <= PROM_BSZ);
    MPI_Isend(proc_gnode_,np+1,MPI_INT,nxpart,tag,prevG->gComm_,&sreq);
    MPI_Wait( &sreq, &status ); 
    // sizes
    int *pb = sbuf; *pb++ = nLocalNd_, *pb++ = nNodes_;
    // send(tag) < nLocalNd_, nNodes_ >
    MPI_Send( sbuf, 2, MPI_INT, nxpart, tag, prevG->gComm_ );
    // send parent updates for last
    pb = sbuf;
    tipos = oldgid_newgid.getHeadPosition(); 
    while( tipos ) {
      gid = oldgid_newgid.getNext( tipos, &oldgid ) - 1;  oldgid--;
      if( pb-sbuf + sendsz >= PROM_BSZ ){
	// send(tag) *<oldgid, newgid>
	MPI_Isend( sbuf,pb-sbuf,MPI_INT, nxpart, tag, prevG->gComm_,&sreq );
	MPI_Wait( &sreq, &status ); 
	pb = sbuf;
      }
      *pb++ = oldgid; *pb++ = gid; 
    }
    // send(tag) *<oldgid, newgid>
    MPI_Isend( sbuf, pb-sbuf,MPI_INT, nxpart, tag, prevG->gComm_, &sreq );
    MPI_Wait( &sreq, &status ); 
    if( pb-sbuf ){      // caper
      MPI_Send( NULL, 0, MPI_INT, nxpart, tag, prevG->gComm_ );
    }
    // send nodes
    for( xx = 0, pb = sbuf ; xx < nNodes_ ; xx++ ){
      PromNode *curr = nodes_[xx]; assert(curr->symbData_ != NULL);
      ierr = curr->GetWrite_size( tt );   CHKERRQ(ierr);
      if( (pb-sbuf) + tt >= PROM_BSZ ) {
	// send(tag) < *node >
	MPI_Isend( sbuf, pb-sbuf,MPI_INT, nxpart, tag, prevG->gComm_, &sreq );
	MPI_Wait( &sreq, &status ); 
	assert(pb-sbuf <= PROM_BSZ);	pb = sbuf; 
      }
      ierr = curr->Write( &pb );   CHKERRQ( ierr );
    }
    if( (pb-sbuf) > 0 ) {
      // send(tag) < *node >
      MPI_Isend( sbuf, pb-sbuf, MPI_INT, nxpart, tag, prevG->gComm_, &sreq );
      MPI_Wait( &sreq, &status ); 
    }
  }
  else{
    assert( prevG->isActive() );
    // recv/send proc_gnod
    MPI_Recv(rbuf,np+1,MPI_INT, prevpart, tag, prevG->gComm_,&status);
    MPI_Get_count( &status, MPI_INT, &tt );    assert(tt == np+1);
    if( nxpart != endpart ){
      // send(tag) proc_gnode
      MPI_Isend(rbuf,np+1,MPI_INT,nxpart,tag,prevG->gComm_,&sreq);
    }
    assert(proc_gnode_ != NULL);
    for( ii = 0 ; ii < np+1 ; ii++  ) proc_gnode_[ii] = rbuf[ii];
    // delete last nodes
    ghosts_.removeAll();
    if( nodes_ != NULL ){ 
      for( xx = 0 ; xx < nNodes_ ; xx++ ) Kill( nodes_[xx] );
      PetscFree( nodes_ ); nodes_ = NULL; 
      nNodes_ = 0; // hack!
    }
    if( nxpart != endpart ) MPI_Wait( &sreq, &status );
    // recieve/send sizes < nLocalNd_, nNodes_ >
    MPI_Recv(rbuf, 2, MPI_INT, prevpart, tag, prevG->gComm_, &status);
    MPI_Get_count( &status, MPI_INT, &tt );    assert(tt == 2);
    if( nxpart != endpart ){
      // send(tag) < nLocalNd_, nNodes_ >
      MPI_Isend( rbuf, 2, MPI_INT, nxpart, tag, prevG->gComm_, &sreq);
    }
    int *pb = rbuf; nLocalNd_ = *pb++; nNodes_ = *pb++;
    // new node array
    assert( nNodes_ > 0 && nodes_ == NULL); // faking out addNNodes()!!!
    ierr = PetscMalloc(sizeof(PromNode*)*nNodes_,&nodes_); CHKERRQ(ierr);
    // oldgid_newgid
    while( 1 ){
      if( nxpart != endpart ) MPI_Wait( &sreq, &status );
      // recv(tag) *<oldgid, newgid>
      MPI_Recv(rbuf, PROM_BSZ, MPI_INT, prevpart, tag, prevG->gComm_, &status);
      MPI_Get_count( &status, MPI_INT, &tt );
      if( nxpart != endpart ){
	// send(tag) *<oldgid, newgid>
	MPI_Isend( rbuf, tt, MPI_INT, nxpart, tag, prevG->gComm_, &sreq );
      }
      if( tt == 0 ) break;  
      for ( pb = rbuf ; pb - rbuf < tt ; ){
	int oldgid = *pb++, newgid = *pb++;
	int add = oldgid_newgid.add( newgid+1, oldgid+1 ); assert(add==1);
      }
      assert(pb-rbuf == tt);
    }

    // recv nodes
    for( xx = 0 ; xx < nNodes_ ; ){ 
      if( nxpart != endpart ) MPI_Wait( &sreq, &status );
      MPI_Recv(rbuf, PROM_BSZ, MPI_INT, prevpart, tag, prevG->gComm_, &status);
      MPI_Get_count( &status, MPI_INT, &tt );
      if( nxpart != endpart ){
	// send(tag) *< node >
	MPI_Isend( rbuf, tt, MPI_INT, nxpart, tag, prevG->gComm_, &sreq );
      }
      pb = rbuf;
      while(pb-rbuf < tt) {
	PromNode *curr = nodes_[xx] = newNode(xx);  assert(!curr->isDead());
	ierr = curr->Read( parBank_, &pb );  CHKERRQ(ierr);  
	assert(pb-rbuf <= tt); 
	if( xx >= nLocalNd_ ){
	  gid = curr->index_.global();  assert(curr->symbData_ != NULL);
	  ierr = GetProc( gid, proc ); CHKERRQ(ierr); 
	  curr->symbData_->ghostID.setProc(proc); ii = gid - proc_gnode_[proc];
	  curr->symbData_->ghostID.setLocal( ii ); assert(ii>=0);
	  ii = ghosts_.add( curr, gid + 1 );	  assert(ii == 1);
	}
	else assert( !curr->isGhost() && curr->index_() == xx );
	xx++; assert(xx <= nNodes_);
      }
      assert(pb-rbuf == tt);
    }
    if( nxpart != endpart ) MPI_Wait( &sreq, &status );
  }
  assert(xx==nNodes_);
  MPI_Iprobe(MPI_ANY_SOURCE, tag, prevG->gComm_, &ii, &status );assert(!ii); 

  return 0;
}
  
#undef __FUNCT__  
#define __FUNCT__ "GlobalizeAdjac"
/* PromGrid::GlobalizeAdjac ******************************
 *
 *  - allocate and get most of new data
 *
 *  INPUT:
 *    - mynew0:
 *    - newLocAdj_oldgid:
 *    - newGhsAdj_oldgid:
 *    - oldgid_newgid:  
 *    - ghostoldgid_newlid:  
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::GlobalizeAdjac( const int mynew0,
			      PromList<int> *newLocAdj_oldgid, 
			      PromList<int> *newGhsAdj_oldgid, 
			      PromTable<int> &oldgid_newgid,
			      PromTable<int> &ghostoldgid_newlid )
{
  assert(ghosts_.isEmpty() && isActive());
  
  for( int xx = 0 ; xx < nNodes_ ; xx++ ){
    PromNode *curr = nodes_[xx], *curr2; 
    PromList<int> *tab;   assert(curr->index_()==xx);assert(!curr->isBound());
    assert(curr->getState() == PROM_UNKNOWN);
    if( xx < nLocalNd_ ) tab = &newLocAdj_oldgid[xx];
    else{ 
      tab = &newGhsAdj_oldgid[xx-nLocalNd_];
      int ii = ghosts_.add( curr, curr->index_.global() + 1 ); assert(ii == 1);
    }
    Position pos = tab->getHeadPosition();   //assert(tipos!=NULL); 
    while( pos ) {
      const int oldgid = tab->getNext( pos ) - 1;
      const int gid = oldgid_newgid.find( oldgid+1 ) - 1;
      if( gid >= 0 ) {
	int lid = gid - mynew0;
	if( lid < 0 || lid >= nLocalNd_ ) {
	  lid = ghostoldgid_newlid.find(oldgid+1) - 1;
	  if( lid >= 0 ) curr2 = nodes_[lid];
	  else curr2 = NULL;
	}
	else curr2 = nodes_[lid];
	assert(curr2 != curr);
	if( curr2 ) curr->symbData_->adjac.add( curr2 );
	else assert(curr->isGhost());
      }
      else assert( curr->isGhost() );
    }
  }
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "MoveGrid"
/* PromGrid::MoveGrid ******************************
 *
 *  - allocate and get most of new data
 *
 *  INPUT:
 *    - oldgid_newgid:  (out)
 *    - ghostoldgid_newlid:  (out)
 *    - newproc_gnode:
 *    - lid_part:
 *    - part_c:
 *    - newLocAdj_oldgid: (out)
 *    - newGhsAdj_oldgid: (out)
 *
 *  SIDE EFFECTS:
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::MoveGrid( PromTable<int> &oldgid_newgid,
			PromTable<int> &ghostoldgid_newlid,
			const int newproc_gnode[], 
			const int lid_part[], const int part_c[],
			PromList<int> *&newLocAdj_oldgid, 
			PromList<int> *&newGhsAdj_oldgid )
{ 
  const int   mype=comm_.mype(),np=comm_.np(),mynew0=newproc_gnode[mype]; 
  const int   newnlocal = newproc_gnode[mype+1]-mynew0,tag=getNewTag();
  int         part,locid,tt,xx,ii,nadj,sz,ierr,receive;
  const int   nCommprocs=np/factor_,mycommpart=mype/factor_,tag1=getNewTag();
  PromNode   *curr;  
  TablePos    tipos; Position pos;
  MPI_Status  status;
  PromNode  **newlocnodes = NULL;
  MPI_Request*req; xx = sizeof(shpfloat); ii = sizeof(int);
  const int   disz=xx/ii; tt = sizeof(MPI_Request);
  const int   reqsz = (tt/xx + ((tt%xx == 0) ? 0 : 1)) * disz;
  PromList<MPI_Request*>  rmessages, smessages, smessages2, buflist;
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 

  if( newnlocal ){
    ierr = PetscMalloc(sizeof(PromNode*)*newnlocal,&newlocnodes);
    CHKERRQ(ierr);
    for( ii = 0 ; ii < newnlocal ; ii++ ) newlocnodes[ii] = NULL;
    newLocAdj_oldgid = new PromList<int>[newnlocal];
    CHKNEW_void(newLocAdj_oldgid,newnlocal);
  }
  else{
    newLocAdj_oldgid = NULL;
    PromContext::printArr(stderr,newproc_gnode,np,"ERROR: empty proc: newproc_gnode[%d]");
    //PetscSleep(10);
    SETERRQ(PETSC_COMM_SELF,1,"This should not happen - use faster coarsening");
  }

  // count sends, alloc messages (tag)
  assert(nCommprocs==np/factor_);
  for( ii = 0 ; ii < nCommprocs ; ii++ ) sbuf[ii] = 0;
  for( part = 0 ; part < nCommprocs ; part++ ){
    if( part == mycommpart ) continue;
    if( part_c[part] ) {
      for( xx = ii = 0, sz = 1 ; xx < nLocalNd_ ; xx++ ){ 
	if( lid_part[xx] == part ){
	  PromNode *curr = nodes_[xx];
	  nadj = curr->symbData_->adjac.getCount();
	  if(sz%disz != 0) sz++; // make even
	  ierr = curr->GetWrite_size(tt);   CHKERRQ(ierr);
	  sz += tt + nadj + 1;
	  ii++; // count
	}
      }
    }
    else ii = sz = 0;
    // alloc send send(tag) 
    if( ii > 0 ) {
      assert(sbuf[part] == 0); sbuf[part] = sz;
      tt = sz + reqsz; assert(tt >= 3 + 2*ii);
      ierr = PetscMalloc( (tt+1)*sizeof(int), &req );           CHKERRQ(ierr);
      smessages.addTail( req ); 
      int *buff = (int*)req; buff[tt-1] = -4; buff[tt] = -part; // sentinal
      buff[reqsz] = ii; // scounts.addTail( ii );  
    }
  }
  // global! 
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, gComm_ );
  // post receives (tag)
  for( ii = 0 ; ii < nCommprocs ; ii++ ) {
    if( rbuf[ii] != 0 ) {
      // recv(tag)
      sz = rbuf[ii];      part = ii;
      ierr = PetscMalloc((sz+reqsz)*sizeof(int),&req);  CHKERRQ(ierr);
      rmessages.addTail( req );      
      int *pb = (int*)req + reqsz;
      // irecv(tag) <count,count*<node,nadj,nadj*<adjac>>>
      MPI_Irecv( pb, sz, MPI_INT, part, tag, gComm_, req );
    }
  }
  // send data (tag)
  pos = smessages.getHeadPosition(); 
  for( part = 0 ; part < nCommprocs ; part++ ){
    if( part_c[part] == 0 || part == mycommpart ) continue;
    req = smessages.getNext( pos ); 
    int *buff = (int*)req + reqsz, *pb = buff, nn = *pb++; // count
    for( xx = tt = 0 ; xx < nLocalNd_ ; xx++ ){
      if( lid_part[xx] == part ) {
	tt++;  // count
	PromNode *curr = nodes_[xx];        assert( !curr->isMarked() ); 
	if((pb-buff)%disz != 0) *pb++ = -3; // make even   
	ierr = curr->Write( &pb ); CHKERRQ(ierr);
	*pb++ = curr->symbData_->adjac.getCount(); 
	TablePos tpos = curr->symbData_->adjac.getHeadPosition();     
	while( tpos ) {
	  PromNode *curr2 = curr->symbData_->adjac.getNext( tpos );
	  *pb++ = curr2->index_.global(); // old gids
	}
      }
    } 
    assert(buff[0]==tt && tt==nn); assert(tt>0 && *pb==-part && *(pb-1)!=-4);
    // send(tag) <count,*<[-3,]node,nadj,nadj*<adjac>>>
    MPI_Isend( buff, pb-buff, MPI_INT, part, tag, gComm_, req );
  }
  assert(!pos);

  // copy nodes in, set (lid, gid), put in old gids in adjacs for stays
  locid = 0;
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ){ 
    if( lid_part[xx] == mycommpart ) {
      curr = newlocnodes[locid] = nodes_[xx]; nodes_[xx] = NULL;  // take it
      const int newgid = mynew0 + locid, oldgid = curr->index_.global();
      ii = oldgid_newgid.add( newgid+1, oldgid+1 ); assert(ii == 1);
      // adjacs
      PromList<int> *adj = &newLocAdj_oldgid[locid]; 
      PromTable<PromNode*> *cadj = &curr->symbData_->adjac;
      while( !cadj->isEmpty() ) {
	PromNode *curr2 = cadj->removeHead();
	adj->addTail(curr2->index_.global()+1); // temp old gid
      }
      assert( !curr->isMarked() ); curr->mark();
      curr->index_ = locid++; 
      if( curr->symbData_->dependents ){ // clean old data
	while( !curr->symbData_->dependents->isEmpty() ) {
	  delete curr->symbData_->dependents->removeHead();
	}
	delete curr->symbData_->dependents; curr->symbData_->dependents = NULL;
      }
      curr->symbData_->ghostID.reset();
      curr->symbData_->adjac.removeAll();
      curr->symbData_->elemList.removeAll();
      curr->symbData_->child = NULL;   assert(curr->symbData_->parent == NULL);
    }
  }
  // receive(tag) nodes & new ghosts(adjacs) (added as gids), 
  assert( smessages2.isEmpty() /*&& scounts.isEmpty()*/ );
  while( !rmessages.isEmpty() ) {
    req = rmessages.removeHead(); assert(*req!=MPI_REQUEST_NULL);
    // recv(tag) <count,*<node,nadj,nadj*<adjac>>>
    MPI_Test( req, &receive, &status );
    if( !receive ) rmessages.addTail( req );
    else {
      int *buff = (int*)req, *pp0 = buff + reqsz, *pb = pp0, *spb = pb;
      int num = *pb++;  assert(num>0); // read and write to this
      *spb++ = num; // keep count 
      while( num-- ){
	assert( newlocnodes[locid] == NULL );
	curr = newlocnodes[locid] = newNode( locid );
	// make even
	if((pb-pp0)%disz != 0){ int n3 = *pb++; assert(n3==-3);} // make even
	ierr = curr->Read( parBank_, &pb );  CHKERRQ(ierr);
	int oldgid = curr->index_.global();
	// new data
	int gid = mynew0 + locid; 
	ii = oldgid_newgid.add( gid+1, oldgid+1 ); assert(ii==1);//
	// adjacs
	nadj = *pb++;
	PromList<int> *adj = &newLocAdj_oldgid[locid]; 
	while(nadj--){ tt = *pb++; adj->addTail(tt+1);} // temp add old gid
	// copy over buffer with...
	*spb++ = locid; *spb++ = oldgid; assert(spb <= pb);//overwite buffer!
	// 
	curr->index_ = locid++;
	assert(!curr->isMarked()); curr->mark();
      }
      MPI_Get_count( &status, MPI_INT, &sz );          assert(pb-pp0 == sz);
      *spb++ = status.MPI_SOURCE; *spb++ = -11;
      // send(tag1) count, count*<newlid,oldgid>, proc, -11
      smessages2.addTail( req ); // save for latter
    }
  }
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &ii, &status );assert(!ii);  
  assert(locid == newnlocal); assert(oldgid_newgid.getCount() == newnlocal);
  // wait for sends (tag), post recv (tag1)
  for( part = 0 ; part < nCommprocs ; part++ ){
    if( part_c[part] == 0 || part == mycommpart ) continue;
    req = smessages.removeHead(); int *buff = (int*)req, *pb = buff + reqsz;
    int num = *pb;/*scounts.removeHead();*/ assert(num>0);
    assert(*req != MPI_REQUEST_NULL);
    MPI_Wait( req, &status ); // do not delete as used latter
    // recv(tag1) count, count*<newlid,oldgid>, pe, -11
    MPI_Irecv( pb, 2*num+3, MPI_INT, part, tag1, gComm_, req );
    buflist.addTail( req );
    pb[2*num+3] = -6;
  }
  // reorder and send(tag1)  <count, count*<newlid,oldgid>,pe,-11>
  if( newnlocal ) {
    // reorder -- pick lowest degree start for Cuthill-McKey ordering
    int *order,oldgid1,newgid;
    ierr = PetscMalloc(newnlocal*sizeof(int),&order);CHKERRQ(ierr);
    for( ii = 0 ; ii < newnlocal ; ii++ ) order[ii] = -1;
    PromList<int> roots; 
    for( locid = 0 ; locid < newnlocal ;/*void*/){
      for( ii = 0, xx = 999999, tt = -1 ; ii < newnlocal ; ii++ ){
	if( order[ii] == -1 ) {
	  if( (sz=newLocAdj_oldgid[ii].getCount() ) < xx){ xx = sz; tt = ii; }
	}
      }
      assert(tt!=-1); order[tt] = locid++; roots.addTail(tt+1); xx = 0; 
      while( !roots.isEmpty() && xx++ < 2000 ){
	tt = roots.removeHead() - 1;         assert( order[tt] != -1 );
	PromList<int> *adj = &newLocAdj_oldgid[tt];
	pos = adj->getHeadPosition(); 
	while( pos ){
	  oldgid1 = adj->getNext(pos); 
	  newgid = oldgid_newgid.find(oldgid1) - 1;
	  if( newgid >= 0 ) {
	    tt = newgid - mynew0; assert(tt >= 0 && tt < newnlocal); 
	    if(order[tt] == -1){ order[tt] = locid++; roots.addTail(tt+1);}
	  }
	}
      }
      while( !roots.isEmpty() ) roots.removeHead(); 
    }
    // set 'newLocAdj_oldgid' 'newlocnodes'
    PromNode **newlocnodes2 = NULL;
    ierr = PetscMalloc( sizeof(PromNode*)*newnlocal, &newlocnodes2 );
    CHKERRQ(ierr);
    for(ii=0;ii<newnlocal;ii++)newlocnodes2[ii]=NULL;
    PromList<int> *temp_adjacs; 
    temp_adjacs = new PromList<int>[newnlocal]; 
    CHKNEW_void(temp_adjacs,newnlocal);
    for(ii=0;ii<newnlocal;ii++){
      int newlid = order[ii]; assert(newlocnodes2[newlid] == NULL);
      PromNode *curr = newlocnodes2[newlid] = newlocnodes[ii];
      curr->index_ = newlid;
      PromList<int> *sadj = &newLocAdj_oldgid[ii]; 
      PromList<int> *taj = &temp_adjacs[newlid]; assert(taj->isEmpty());
      while( !sadj->isEmpty() ) taj->addTail( sadj->removeHead() );       
    }
    PromNode **ttemp = newlocnodes;
    newlocnodes = newlocnodes2; PetscFree( ttemp );
    PromList<int> *tttemp = newLocAdj_oldgid;  
    newLocAdj_oldgid = temp_adjacs;
    delete [] tttemp; tttemp = temp_adjacs = NULL;
    // set new 'oldgid_newgid'
    PromList<int> oldnewgids(2*newnlocal);
    tipos = oldgid_newgid.getHeadPosition(); 
    while( tipos ) {
      int oldgid1, newgid, newnewgid, lid;
      newgid = oldgid_newgid.getNext( tipos, &oldgid1 ) - 1;
      lid = newgid - mynew0; assert(lid>=0 && lid<newnlocal);
      newnewgid = order[lid] + mynew0;
      oldnewgids.addTail(oldgid1); oldnewgids.addTail(newnewgid+1); 
    }
    oldgid_newgid.removeAll();
    while( !oldnewgids.isEmpty() ) {
      int oldgid1=oldnewgids.removeHead(),newgid1=oldnewgids.removeHead();
      ii = oldgid_newgid.add( newgid1, oldgid1 ); assert(ii==1);
    }
    // send(tag1)  <count, count*<newlid,oldgid>,pe,-11>
    pos = smessages2.getHeadPosition(); 
    while( pos ){
      req = smessages2.getNext( pos ); 
      int *buff = (int*)req + reqsz, *pb = buff, nn= *pb++;
      // map to new
      for( ii = 0 ; ii < nn ; ii++, pb += 2){
	// <newlid,oldgid>
	int lid = *pb; assert(lid>=0 && lid<newnlocal);
	*pb =  order[lid];
      }
    }
    PetscFree( order );
  }
  // send(tag1)
  pos = smessages2.getHeadPosition(); 
  while( pos ){
    req = smessages2.getNext( pos ); 
    int *buff = (int*)req + reqsz, *pb = buff, nn = *pb++, proc; 
    pb += 2*nn;
    // send(tag1) <count, count*<newlid,oldgid>,pe,-11>
    proc = *pb++;  assert(*pb == -11);
    MPI_Isend( buff, 2*nn+3, MPI_INT, proc, tag1, gComm_,req );
  }
  // count ghosts, give ids
  int ghost_lid = newnlocal; 
  for( xx = 0 ; xx < newnlocal ; xx++ ){ 
    Position pos = newLocAdj_oldgid[xx].getHeadPosition();     
    while( pos ) {
      int oldgid = newLocAdj_oldgid[xx].getNext(pos) - 1;
      // if neigh. not local and not added yet - add 
      if(!oldgid_newgid.find(oldgid+1) && !ghostoldgid_newlid.find(oldgid+1)){
	int a = ghostoldgid_newlid.add(ghost_lid+1,oldgid+1); assert(a>0);
	ghost_lid++;
      }
    }
  }
  const int numghosts = ghostoldgid_newlid.getCount();
  if( numghosts ){
    newGhsAdj_oldgid = new PromList<int>[numghosts];
    CHKNEW_void(newGhsAdj_oldgid,numghosts);
  }
  else newGhsAdj_oldgid = NULL; // one proc
  const int newnnodes = numghosts + newnlocal; assert(newnnodes<=nglobal());
  // make new node arrays
  PromNode **newnodes = NULL;
  if( newnnodes ) { 
    ierr = PetscMalloc(sizeof(PromNode*)*newnnodes,&newnodes);
    CHKERRQ(ierr); // marks?????
    for( xx = 0; xx < newnlocal ; xx++ ) {
      newnodes[xx] = newlocnodes[xx];      assert(newnodes[xx]!=NULL);
    }
    for( ; xx < newnnodes ; xx++ ) newnodes[xx] = NULL;
    PetscFree( newlocnodes ); newlocnodes = NULL;    
  }
  else{ assert(newlocnodes == NULL); assert(0); }

  // recv(tag1) < count, count*<newlid,oldgid>, pe, -11 > of old nodes
  while( !buflist.isEmpty() ) {
    req = buflist.removeHead(); assert(*req != MPI_REQUEST_NULL);
    MPI_Test( req, &receive, &status );
    if( !receive ) buflist.addTail( req );
    else {
      int *buff = (int*)req, *pb = buff + reqsz;
      MPI_Get_count( &status, MPI_INT, &sz );  assert(sz%2==1);
      part = status.MPI_SOURCE;
      // recv(tag1) count, count*< newlid, oldgid >, pe, -11
      ii = (sz-3)/2; xx = *pb++; assert(xx == ii);
      int partgid0 = newproc_gnode[part*factor_];
      while( ii-- ) {
	int newlid = *pb++, newgid = partgid0 + newlid, oldgid = *pb++;
	int a = oldgid_newgid.add( newgid+1, oldgid+1 );
	assert(a == 1 || -a == oldgid+1);
      }
      xx = *pb++; ii = *pb++; 
      assert(*pb == -6); assert(xx == mype/factor_); assert(ii == -11); 
      PetscFree( buff ); 
    }
  }
  MPI_Iprobe(MPI_ANY_SOURCE, tag1, gComm_, &ii, &status );assert(!ii); 
  // wait for sends(tag1)
  while( !smessages2.isEmpty() ) {
    req = smessages2.removeHead(); assert(*req!=MPI_REQUEST_NULL);
    MPI_Test( req, &receive, &status );
    if( !receive ) rmessages.addTail( req );
    else PetscFree(req);
  }
  // make ghosts -- lid
  tipos = ghostoldgid_newlid.getHeadPosition(); 
  while( tipos ) {
    int oldgid, lid;
    lid = ghostoldgid_newlid.getNext( tipos, &oldgid ) - 1; oldgid--;
    assert( lid >= newnlocal && lid < newnnodes );
    assert( newnodes[lid] == NULL ); 
    curr = newnodes[lid] = newNode( lid ); 
    assert( !curr->isGhost() && curr->index_() == lid );
    curr->index_.setGlobal( oldgid ); // old gids on ghosts!!!
  }
  // ghosts -- find local pointers to some
  for( xx = 0; xx < nLocalNd_ ; xx++ ){
    if( nodes_[xx] != NULL ) {
      PromNode *oldc = nodes_[xx];
      int oldgid = oldc->index_.global(); 
      // if new ghost, and have a gid (else doit latter)
      int lid = ghostoldgid_newlid.find( oldgid+1 ) - 1;
      if( lid > 0 && oldgid_newgid.find( oldgid+1 ) ) {
	assert( newnodes[lid] != NULL );
	assert( lid >= newnlocal && lid < newnnodes ); // a ghost
	PromList<int> *adj = &newGhsAdj_oldgid[lid-newnlocal];
	assert(adj!=NULL);
	while( !oldc->symbData_->adjac.isEmpty() ) {
	  PromNode *curr2 = oldc->symbData_->adjac.removeHead();
	  adj->addTail( curr2->index_.global()+1 ); //temp old gid
	}
	PromNode *curr = newnodes[lid];
	curr->index_ = lid; 
	assert( !curr->isMarked() );  curr->mark(); // have adjacs
      }
      else if( lid > 0 ) {
	assert( newnodes[lid] && newnodes[lid]->nParents() == 0 );
      }
    }
  } 
  // set new
  for(xx=0; xx<nNodes_;xx++) if(nodes_[xx] != NULL) Kill( nodes_[xx] );
  PetscFree( nodes_ );
  nLocalNd_ = newnlocal; nNodes_ = numghosts + newnlocal;
  nodes_ = newnodes;  
  ghosts_.removeAll();
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &ii, &status );assert(!ii); 
  MPI_Iprobe(MPI_ANY_SOURCE, tag1, gComm_, &ii, &status );assert(!ii); 

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "CollectGhostData"
/* PromGrid::CollectGhostData ******************************
 *
 *  - 
 *
 *  INPUT:
 *    - oldgid_newgid:  
 *    - ghostoldgid_newlid:
 *    - oldproc_gnode:
 *    - lid_part[oldproc_gnode[mype+factor_]-oldproc_gnode[mype]]:
 *    - newLocAdj_oldgid[]:
 *    - newGhsAdj_oldgid[]: (in/out)
 *
 *  SIDE EFFECTS:
 *    - puts data in ghost nodes
 *    - marks all ghost nodes
 *
 *  RETURN: 
 *     - PETSc error code
 */
int PromGrid::CollectGhostData(PromTable<int> &oldgid_newgid, 
			       PromTable<int> &ghostoldgid_newlid,
			       const int oldproc_gnode[], 
			       const int lid_part[], 
			       const PromList<int> newLocAdj_oldgid[], 
			       PromList<int> newGhsAdj_oldgid[] )
{
  int        ierr,tt,xx,ii,lid,olid,oldgid,gid,proc;
  const int tag = getNewTag(),tag1=getNewTag(),tag2=getNewTag();
  const int  mype=comm_.mype(),np=comm_.np();
  const int  mycommpart = mype/factor_; assert(proc_gnode_[0]==0);
  const int  mynew0 = proc_gnode_[mype], myold0 = oldproc_gnode[mype];
  MPI_Status status; TablePos tpos;
  PromNode  *ghos; tt = sizeof(MPI_Request); xx = sizeof(shpfloat);
  const int  gsz = xx/sizeof(int);  assert(oldproc_gnode[np]==proc_gnode_[np]);
  const int  reqsz = (tt/xx + ((tt%xx == 0) ? 0 : 1)) * gsz;
  PromList<MPI_Request*> smessages, rmessages;
  PromTable<PromList<int>*> proc_list1(57), proc_list2(57);
  PromList<int>  *list, slist2(200); 
  
  // count todos, get rest of ghost data -- send out requests
  int todo = 0; assert(ghostoldgid_newlid.getCount() == nNodes_-nLocalNd_);
  tpos = ghostoldgid_newlid.getHeadPosition();  
  while( tpos ) {
    xx = ghostoldgid_newlid.getNext( tpos, &oldgid ) - 1; oldgid--;
    ghos = nodes_[xx];          assert(xx >= nLocalNd_ && xx < nNodes_);
    //
    if( !ghos->isMarked() ) {
      assert( ghos->getState() == PROM_UNKNOWN );
      ierr = FindProc( oldproc_gnode, np, oldgid, proc ); CHKERRQ(ierr);
      assert( proc != mype );
      if( (list=proc_list1.find(proc+1)) == NULL ) {
	list = new PromList<int>(100); CHKNEW(list,1);
	proc_list1.add( list, proc+1 );
      }
      list->addTail( oldgid + 1 ); 
      list->addTail( xx + 1 );
      todo++;
    }
    else { // have it - set as ghost 
      gid = oldgid_newgid.find(ghos->index_.global()+1) - 1;  assert(gid>=0);
      ghos->index_.setGlobal( gid );                   // new gids
      ierr = GetProc( gid, proc ); 
      CHKERRQ(ierr); 
      ii = gid - proc_gnode_[proc];
      ghos->symbData_->ghostID.setProc( proc );     assert(proc != mype);
      ghos->symbData_->ghostID.setLocal( ii );      assert(ghos->isGhost());
      // need geq!
      if( (list=proc_list2.find(proc+1)) == NULL ) {
	list = new PromList<int>(100); CHKNEW(list,1);
	proc_list2.add( list, proc+1 );
      }
      list->addTail( oldgid + 1 );
      list->addTail( xx + 1 ); 
      list->addTail( mype/factor_ + 1 );
      todo++;
    }
  }

  // count sends of requests
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 
  for( ii = 0 ; ii < np/factor_ ; ii++ ) sbuf[ii] = 0;
  tpos = proc_list1.getHeadPosition(); 
  while( tpos ) {
    list = proc_list1.getNext( tpos, &proc ); proc--; 
    tt = list->getCount();
    assert(sbuf[proc/factor_] == 0); sbuf[proc/factor_] = tt;
  }

  // global!
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, gComm_ );
  // post recv(tag) of requests
  for( ii = 0 ; ii < np/factor_ ; ii++ ) {
    if( rbuf[ii] != 0 ) {
      // recv(tag)
      int sz = rbuf[ii], part = ii;
      int *pb; ierr = PetscMalloc((sz+reqsz)*sizeof(int),&pb); CHKERRQ(ierr);
      MPI_Request *rreq = (MPI_Request*)pb; pb += reqsz;
      // irecv(tag) *< oldgid, loclid >
      MPI_Irecv( pb, sz, MPI_INT, part, tag, gComm_, rreq );
      rmessages.addTail( rreq );
    }
  }

  // send(tag) requests
  while( !proc_list1.isEmpty() ) {
    list = proc_list1.removeHead( &proc ); proc--;
    tt = list->getCount();
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);   smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) *pb++ = list->removeHead() - 1; 
    assert(tt == pb-buff);
    // send(tag) *< oldgid, loclid >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq );
    delete list; 
  }

  // recv(tag) requests, collect pass ons
  while( !rmessages.isEmpty() ) {
    MPI_Request *req = rmessages.removeHead(); 
    assert(*req != MPI_REQUEST_NULL);
    // recv(tag) * < oldgid, loclid >
    MPI_Test( req, &ii, &status );
    if( !ii ){ rmessages.addTail( req ); continue; }

    MPI_Get_count( &status, MPI_INT, &tt );     assert(tt > 0 && tt%2 == 0);
    int *buff = (int*)req + reqsz, *pb = buff;
    int part = status.MPI_SOURCE; 
    while( pb-buff < tt ) {
      oldgid = *pb++; olid = *pb++; 
      int part2 = lid_part[oldgid - myold0];     // owner
      if( part2 == mycommpart ) { // I'm real owner -- send right back
	slist2.addTail(oldgid+1);slist2.addTail(olid+1);slist2.addTail(part+1);
      }
      else { // pass on
	proc = part2*factor_;
	if( (list=proc_list2.find(proc+1)) == NULL ) {
	  list = new PromList<int>(100); CHKNEW(list,1);
	  proc_list2.add( list, proc+1 );
	}
	list->addTail(oldgid+1); 
	list->addTail(olid+1); 
	list->addTail(part+1);
      }
    }
    assert(pb-buff == tt);      PetscFree(req);
  }

  // count sends of pass ons
  for( ii = 0 ; ii < np/factor_ ; ii++ ) sbuf[ii] = 0;
  tpos = proc_list2.getHeadPosition(); 
  while( tpos ) {
    list = proc_list2.getNext( tpos, &proc ); proc--;   
    tt = list->getCount();
    assert(sbuf[proc/factor_] == 0); sbuf[proc/factor_] = tt;
  } 

  // global!
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, gComm_ );
  // post recv(tag1) for pass ons
  for( ii = 0 ; ii < np/factor_ ; ii++ ) {
    if( rbuf[ii] != 0 ) {
      // recv(tag1)
      int sz = rbuf[ii], part = ii;
      MPI_Request *rreq; ierr = PetscMalloc((sz+reqsz)*sizeof(int),&rreq);
      CHKERRQ(ierr); 
      int *pb = (int*)rreq + reqsz;
      // irecv(tag1) *< oldgid, lid, client part >
      MPI_Irecv( pb, sz, MPI_INT, part, tag1, gComm_, rreq );
      rmessages.addTail( rreq );
    }
  }

  // send(tag1) pass ons
  while( !proc_list2.isEmpty() ) {
    list = proc_list2.removeHead( &proc ); proc--;
    tt = list->getCount();
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) *pb++ = list->removeHead() - 1;
    assert(tt == pb-buff);
    // send(tag1) *< oldgid, lid, client part >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag1, gComm_, sreq );
    delete list;
  }

  // recv(tag1) pass ons
  while( !rmessages.isEmpty() ) {
    MPI_Request *req = rmessages.removeHead(); 
    assert(*req != MPI_REQUEST_NULL);
    // recv(tag1) *< oldgid, otherlocid, part > 
    MPI_Test( req, &ii, &status );
    if( !ii ){ rmessages.addTail( req ); continue; }
    MPI_Get_count( &status, MPI_INT, &tt );     assert(tt > 0 && tt%3 == 0);
    int *buff = (int*)req + reqsz, *pb = buff;
    while( pb-buff < tt ) {
      slist2.addTail( *pb++ + 1 ); //oldgid
      slist2.addTail( *pb++ + 1 ); //olid
      slist2.addTail( *pb++ + 1 ); //part
    }
    assert(pb-buff == tt);      PetscFree(req); 
  }

  // collect send(tag2) of data
  while( !slist2.isEmpty() ) {  
    oldgid = slist2.removeHead()-1; 
    olid = slist2.removeHead()-1;
    proc = (slist2.removeHead()-1)*factor_;
    if( (list=proc_list1.find(proc+1)) == NULL ) {
      list = new PromList<int>(100); CHKNEW(list,1);
      proc_list1.add(list, proc+1);
    }
    list->addTail(oldgid+1); list->addTail(olid+1); 
  }

  // send(tag2)
  while( !proc_list1.isEmpty() ) {
    list = proc_list1.removeHead( &proc ); proc--;
    // count
    Position pos = list->getHeadPosition();     tt = 0;
    while( pos ){ 
      oldgid = list->getNext(pos) - 1; olid = list->getNext(pos) - 1;
      gid = oldgid_newgid.find( oldgid+1 ) - 1;   assert(gid>=0);
      lid = gid - mynew0; assert(lid>=0); assert(lid<nLocalNd_);
      const PromNode *curr = nodes_[lid]; 
      // count
      if( tt%gsz != 0 ) tt++;
      ierr = curr->GetWrite_size( ii );  CHKERRQ(ierr);
      tt += 4 + ii + newLocAdj_oldgid[lid].getCount();
    }
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) { 
      oldgid = list->removeHead()-1; olid = list->removeHead()-1;
      // 
      gid = oldgid_newgid.find( oldgid+1 ) - 1; assert(gid>=0);
      lid = gid - mynew0; assert(lid>=0 && lid<nLocalNd_);
      const PromNode *curr = nodes_[lid]; 
      // send back ans
      *pb++ = olid; *pb++ = oldgid; 
      if( (pb-buff)%gsz != 0 ) *pb++ = -3;
      ierr = curr->Write( &pb ); CHKERRQ(ierr);
      *pb++ = curr->index_(); // not sent in 'Write'
      *pb++ = newLocAdj_oldgid[lid].getCount(); 
      Position pos = newLocAdj_oldgid[lid].getHeadPosition(); assert(pos!=0);
      while(pos)gid=*pb++=newLocAdj_oldgid[lid].getNext(pos)-1; //old gids
    }
    assert(tt == pb-buff);
    // sends(tag2) *< lid, oldgid, node, glid, numadjac, *adj >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag2, gComm_, sreq );
    delete list;
  }
  // receive data
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag2, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); proc = status.MPI_SOURCE*factor_;
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag2) *< lid, oldgid, node, glid, numadjac, *adj >  
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag2, gComm_, &status );
    while(pb-buff < tt) {
      todo--; assert(todo>=0);
      // add it 
      xx = *pb++; oldgid = *pb++; assert(xx>=nLocalNd_ && xx<nNodes_);
      ghos = nodes_[xx];            
      // new one 
      if( (pb-buff)%gsz != 0 ){ ii = *pb++; assert(ii == -3); }
      ierr = ghos->Read( parBank_, &pb );  CHKERRQ(ierr);
      ii = *pb++; ghos->symbData_->ghostID.setLocal( ii );
      gid = ghos->index_.global(); // new gid
      ghos->symbData_->ghostID.setProc( proc ); // ghost
      int add = oldgid_newgid.add( gid+1, oldgid+1 );
      int nadj = *pb++; assert(nadj>0);  ghos->mark();
      PromList<int> *tab = &newGhsAdj_oldgid[xx-nLocalNd_];
      assert( tab->isEmpty() || add == 0 );
      while( nadj-- ){ // old gids
	gid = *pb++;
	if( add > 0 ) tab->addTail(gid+1); 
      } 
    }
    assert(pb-buff == tt);    PetscFree(buff);
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &ii, &status );assert(!ii);
  MPI_Iprobe(MPI_ANY_SOURCE, tag1, gComm_, &ii, &status );assert(!ii);
  MPI_Iprobe(MPI_ANY_SOURCE, tag2, gComm_, &ii, &status );assert(!ii);

  // unmark
  for( xx = 0; xx < nNodes_ ; xx++ ){
    ghos = nodes_[xx];  assert( ghos->isMarked() ); ghos->mark( FALSE ); 
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "FixLastGrid"
/* PromGrid::FixLastGrid ******************************
 *
 *  - set new (non-trivial) parents < proc, gid > up to here -- and
 *    setup dummy grids for last actives 
 *
 *  INPUT:
 *    - oldgid_newgid:  
 *    - oldproc_gnode[np+1]:
 *    - lid_part[nLocalOld]
 *    - lastG:
 * 
 *  SIDE EFFECTS:
 *    - puts data in ghost nodes
 *
 *  RETURN:
 *     - PETSc error code
 */
int PromGrid::FixLastGrid( PromTable<int> &oldgid_newgid, 
			   const int oldproc_gnode[], const int lid_part[], 
			   PromGrid *prevG )
{
  int ierr,tt,xx,ii,lid,gid,proc,proc2,todo=0,oldgid;
  const int tag=prevG->getNewTag(),tag1=prevG->getNewTag(),tag2=prevG->getNewTag();
  const int mype=comm_.mype(),np=comm_.np();
  const int masterproc = (mype/factor_)*factor_;
  const int lastfactor = prevG->factor_,myold0 = oldproc_gnode[masterproc];   
  MPI_Status status;                          assert(prevG&&prevG->isActive());
  TablePos tpos;
  const int mynew0 = proc_gnode_[masterproc]; xx = sizeof(shpfloat);
  tt = sizeof(MPI_Request);
  const int reqsz = (tt/xx + ((tt%xx == 0) ? 0 : 1)) * (xx/sizeof(int));
  PromList<int> *list2, *list, slist2(200); 
  PromList<MPI_Request*> smessages, rmessages; 
  PromNode *fine;   ii = 4*(prevG->nNodes_-prevG->nLocalNd_)/3;
  PromTable<PromList<int>*> oldgid_list_of_pars(ii+17),proc_list(57);
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 

  // for all nodes (cols) on last level
  for( xx = 0 ; xx < prevG->nNodes_ ; xx++ ){
    fine = prevG->nodes_[xx];
    PromParent **parI = fine->parents_->data_, *par;         // global indices
    for( int II = 0 ; II < fine->nParents() ; II++ ) {
      par = parI[II];
      oldgid = par->global(); 
      gid = oldgid_newgid.find(oldgid+1) - 1, lid = gid - mynew0;
      PromNode *curr = NULL;
      if(lid >= 0 && lid < nLocalNd_) { curr = nodes_[lid]; proc = masterproc;}
      else if( gid >= 0 && (curr=ghosts_.find(gid+1)) ) {
	proc = curr->symbData_->ghostID.proc();
      }
      if( curr != NULL ) {
	par->setGlobal( gid );  par->setProc( proc );
	par->setGEQ( curr->geq() );
	assert(par->isMarked()==FALSE);
      }
      else {
	// get data
	if( !(list=oldgid_list_of_pars.find(oldgid+1)) ){
	  list = new PromList<int>; CHKNEW_void(list,1); 
	  oldgid_list_of_pars.add( list, oldgid+1 );
	  proc = par->proc();
	  if( (list2=proc_list.find(proc+1)) == NULL ) {
	    list2 = new PromList<int>(100); CHKNEW(list2,1);
	    proc_list.add( list2, proc+1 );
	  }
	  list2->addTail(oldgid+1);
	  todo++;
	}
	assert(II < PROM_MAXPAR_HARD);
	int key = PROM_MAXPAR_HARD * xx + II; 
	list->addTail(key+1); 
	par->mark(); // to be done
      }
    }
  }
  // count send(tag)
  for( ii = 0, tt = np/lastfactor ; ii < tt ; ii++ ) sbuf[ii] = 0;
  tpos = proc_list.getHeadPosition(); 
  while( tpos ) {
    list = proc_list.getNext( tpos, &proc ); proc--; 
    tt = list->getCount();
    assert(sbuf[proc/lastfactor] == 0); sbuf[proc/lastfactor] = tt;
  }
  // global!
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, prevG->gComm_ );
  // post recv(tag)
  for( ii = 0 ; ii < np/lastfactor ; ii++ ) {
    if( rbuf[ii] != 0 ) {
      int sz = rbuf[ii], part = ii;
      int *pb; ierr = PetscMalloc((sz+reqsz)*sizeof(int),&pb);   CHKERRQ(ierr);
      MPI_Request *rreq = (MPI_Request*)pb; pb += reqsz;
      // irecv(tag) *< oldgid >
      MPI_Irecv( pb, sz, MPI_INT, part, tag, prevG->gComm_, rreq );
      rmessages.addTail( rreq );
    }
  }
  // send(tag)
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    tt = list->getCount();
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
    CHKERRQ(ierr);    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) { xx = list->removeHead() - 1; *pb++ = xx; }
    assert(tt == pb-buff);
    // send(tag) *< oldgid >
    MPI_Isend( buff, tt, MPI_INT, proc/lastfactor, tag, prevG->gComm_, sreq );
    delete list;
  }

  // receive(tag) dependency message < oldgid > 
  while( !rmessages.isEmpty() ) {
    MPI_Request *req = rmessages.removeHead(); 
    assert(*req != MPI_REQUEST_NULL);
    // recv(tag) * < oldgid >
    MPI_Test( req, &ii, &status );
    if( !ii ){ rmessages.addTail( req ); continue; }
    proc = status.MPI_SOURCE * lastfactor;
    MPI_Get_count( &status, MPI_INT, &tt );     assert(tt > 0);
    int *buff = (int*)req + reqsz, *pb = buff;
    while( pb-buff < tt ) {
      oldgid = *pb++; 
      lid = oldgid - myold0;  assert(lid>=0);
      proc2 = lid_part[lid] * factor_; // master
      if( proc2 == mype ) {
	gid = oldgid_newgid.find( oldgid+1 ) - 1; assert(gid>=0); 
	slist2.addTail(proc+1); slist2.addTail(oldgid+1);
	slist2.addTail(gid+1);
      }
      else {
	if( (list=proc_list.find(proc2+1)) == NULL ) {
	  list = new PromList<int>(100); CHKNEW(list,1);
	  proc_list.add( list, proc2+1 );
	}
	list->addTail(oldgid+1); list->addTail(proc+1); 
      }
    }
    PetscFree(req);     assert(pb-buff == tt);
  }
  // count send(tag1)
  for( ii = 0, tt = np/lastfactor ; ii < tt ; ii++ ) sbuf[ii] = 0;
  tpos = proc_list.getHeadPosition(); 
  while( tpos ) {
    list = proc_list.getNext( tpos, &proc ); proc--; 
    tt = list->getCount(); assert(tt%2==0);
    assert(sbuf[proc/lastfactor] == 0); sbuf[proc/lastfactor] = tt;
  }
  // global!
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, prevG->gComm_ );
  // post recv(tag1)
  for( ii = 0 ; ii < np/lastfactor ; ii++ ) {
    if( rbuf[ii] != 0 ) {
      int sz = rbuf[ii], part = ii;
      int *pb; ierr = PetscMalloc((sz+reqsz)*sizeof(int),&pb); CHKERRQ(ierr);
      MPI_Request *rreq = (MPI_Request*)pb; pb += reqsz;
      // irecv(tag1) *< oldgid, client proc >
      MPI_Irecv( pb, sz, MPI_INT, part, tag1, prevG->gComm_, rreq );
      rmessages.addTail( rreq );
    }
  }
  // send(tag1)
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    tt = list->getCount(); assert(tt%2==0);
    MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz)*sizeof(int),&sreq ); 
    CHKERRQ(ierr);    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) *pb++ = list->removeHead() - 1; 
    assert(tt == pb-buff);
    // send(tag1) *< oldgid, client proc >
    MPI_Isend(buff, tt, MPI_INT, proc/lastfactor, tag1, prevG->gComm_,sreq);
    delete list;
  }
  // receive(tag1) dependency message *< oldgid, client proc > 
  while( !rmessages.isEmpty() ) {
    MPI_Request *req = rmessages.removeHead(); 
    assert(*req != MPI_REQUEST_NULL);
    // recv(tag) * < oldgid, client proc >
    MPI_Test( req, &ii, &status );
    if( !ii ){ rmessages.addTail( req ); continue; }
    MPI_Get_count( &status, MPI_INT, &tt );     assert(tt > 0 && tt%2 == 0);
    int *buff = (int*)req + reqsz, *pb = buff;
    while( pb-buff < tt ) {
      oldgid = *pb++; proc = *pb++; 
      gid = oldgid_newgid.find( oldgid+1 ) - 1;  assert(gid >= 0);
      slist2.addTail(proc+1);slist2.addTail(oldgid+1);slist2.addTail(gid+1);
    }
    PetscFree(req);     assert(pb-buff == tt);
  }
  // collect send(tag2)
  while( !slist2.isEmpty() ) {  
    proc = slist2.removeHead()-1; oldgid = slist2.removeHead()-1; 
    gid = slist2.removeHead()-1;
    if( (list=proc_list.find(proc+1)) == NULL ) {
      list = new PromList<int>(100); CHKNEW(list,1);
      proc_list.add(list, proc+1);
    }
    list->addTail(oldgid+1); list->addTail(gid+1); 
  }
  // send(tag2)
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    // count
    tt = 3*(list->getCount()/2);
    MPI_Request *sreq; ierr = PetscMalloc( (tt+reqsz)*sizeof(int),&sreq ); 
    CHKERRQ(ierr);    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      oldgid = list->removeHead()-1; gid = list->removeHead()-1;
      // send back ans
      *pb++ = oldgid; *pb++ = gid; *pb++ = nodes_[gid-mynew0]->geq(); 
    }
    assert(tt == pb-buff);
    // sends(tag2) *< oldgid, gid, geq >
    MPI_Isend(buff,tt, MPI_INT, proc/lastfactor, tag2, prevG->gComm_, sreq );
    delete list;
  }
  // recv(tag2)
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag2, prevG->gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt );     assert(tt > 0 && tt%3==0);
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // receive(tag2) dependency message *< oldgid, gid, geq > 
    MPI_Recv(buff,tt,MPI_INT,status.MPI_SOURCE,tag2,prevG->gComm_,&status);
    while( pb-buff < tt ) {
      todo--; assert(todo>=0);
      const int oldgid = *pb++, gid = *pb++, geq = *pb++; 
      ierr = GetProc( gid, proc ); CHKERRQ(ierr);
      list = oldgid_list_of_pars.find(oldgid+1);      assert(list!=NULL);
      while( !list->isEmpty() ) {
	ii = list->removeHead()-1; xx = ii/PROM_MAXPAR_HARD; 
	int II = ii%PROM_MAXPAR_HARD;
	PromParent *p = prevG->nodes_[xx]->parents_->data_[II]; 
	assert(p->isMarked());  p->mark(FALSE);
	p->setGlobal( gid );    p->setProc( proc ); p->setGEQ( geq ); 
      }
      delete list;
    }
    assert(pb-buff == tt);    PetscFree(buff);
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 
  MPI_Iprobe( MPI_ANY_SOURCE, tag, prevG->gComm_, &ii, &status );assert(!ii);
  MPI_Iprobe( MPI_ANY_SOURCE, tag1, prevG->gComm_, &ii, &status );assert(!ii);
  MPI_Iprobe( MPI_ANY_SOURCE, tag2, prevG->gComm_, &ii, &status );assert(!ii);

  return 0;  
}

#undef __FUNCT__  
#define __FUNCT__ "Repartition"
/* PromGrid::Repartition *******************************************
 *
 *   - 
 *
 *   INPUT: 
 *     - fnode_part: (out)
 *     - part_c:(out)
 *     - newproc_gnode:(out)
 *
 *   RETURN:
 *     - PETSc error code 
 */
int PromGrid::Repartition( int *&lid_part, int *&part_c, int *&newproc_gnode )
{
  const int np=comm_.np(), mype = comm_.mype(), verb = (options_.verbose_>1);
  int ierr, num, ii, xx, tt, gid;       assert(isActive()); 

  // get local partition
  assert( !lid_part && nLocalNd_ > 0); 
  ierr = PetscMalloc( nLocalNd_*sizeof(int), &lid_part ); CHKERRQ(ierr);
  assert(np > factor_ );

  assert(nLocalNd_ > 0);
  const int nCommprocs = np/factor_;   
  // nodes -> proc data
  idx_t *vtxdist; 
  ierr = PetscMalloc((nCommprocs+1)*sizeof(idx_t),&vtxdist);
  CHKERRQ(ierr);
  for( ii = xx = 0 ; xx < nCommprocs; ii += factor_, xx++ ) {
    vtxdist[xx] = proc_gnode_[ii];
  }
  vtxdist[nCommprocs] = proc_gnode_[np]; assert(ii==np);

  // make partitions
  idx_t *vwgts = NULL, *xadj = NULL;
  ierr = PetscMalloc((nLocalNd_)*sizeof(idx_t),&vwgts);CHKERRQ(ierr);
  ierr = PetscMalloc((nLocalNd_+1)*sizeof(idx_t),&xadj);CHKERRQ(ierr);
  
  // set edge wieghts ('ewgts'), and  'xadj' & 'adjncy' adjacentcy data
  for( xx = num = 0 ; xx < nLocalNd_ ; xx++ ){  
    xadj[xx] = num; 
    tt = nodes_[xx]->symbData_->adjac.getCount();     num += tt; 
  }
  xadj[nLocalNd_] = num; const int numadjac = num; assert(nLocalNd_ == xx);
  MPI_Allreduce( &num, &tt, 1, MPI_INT, MPI_MIN, gComm_ );
  if( tt == 0 ){
    PetscFree(lid_part); PetscFree(vtxdist); // isolated body!!!
    PetscFree(xadj);  
    if(vwgts != NULL) PetscFree(vwgts); 
    return 5;
  }
  // make 'adjncy'
  idx_t *adjncy = NULL;
  ierr = PetscMalloc((numadjac+1)*sizeof(idx_t),&adjncy);
  CHKERRQ(ierr);      adjncy[numadjac] = -1;
  idx_t *ewgts = NULL;
  //ierr = PetscMalloc((numadjac+1)*sizeof(idx_t),&ewgts); CHKERRQ(ierr); 
  //ewgts[numadjac] = -1;
  for( xx = ii = 0 ; xx < nLocalNd_ ; xx++ ){
    PromNode *curr = nodes_[xx];  
    gid = curr->index_.global();
    TablePos tpos = curr->symbData_->adjac.getHeadPosition();  
    while( tpos ) {
      PromNode *curr2 = curr->symbData_->adjac.getNext( tpos );
      int gid2 = curr2->index_.global(); 
      assert( gid2 != gid );
      if(ewgts != NULL) ewgts[ii] = 1;
      adjncy[ii] = gid2; 
      ii++;
    }
    if(vwgts != NULL) vwgts[xx] = 1+curr->symbData_->adjac.getCount()/4;
  }
  assert(numadjac == ii); 
  assert(nCommprocs<=vtxdist[nCommprocs]); 
  
  // call metis for partitioning
  int options[6] = {0,250,1,0,1,0},weightflag=0,nfl=0,ncon=1,npp=nCommprocs;
  if( ewgts != NULL ) weightflag++;
  if( vwgts != NULL) weightflag += 2; 
  tt=numadjac; MPI_Allreduce(&tt,&ii,1,MPI_INT,MPI_SUM,gComm_); // debug
  if( ii%2 != 0 ) { SETERRQ1(PETSC_COMM_SELF,1,"numadjac(%d) != 0",ii); }
  perf_mon_.EventBegin( PromContext::prom_logs_[LOG_METIS] );
  if(verb>0)PetscPrintf(gComm_,"[%d]%s ParMETIS_V3_AdaptiveRepart...\n",mype,__FUNCT__);
  real_t tpwgts[PROM_BSZ], ubvec=1.05, itr=100.;
  for(ii=0;ii<npp;ii++)tpwgts[ii]=1./npp;
  ParMETIS_V3_AdaptiveRepart(vtxdist, xadj, adjncy, vwgts, vwgts, ewgts, 
			     &weightflag, &nfl, &ncon, &npp, tpwgts, &ubvec, 
			     &itr, options, &ii, lid_part, &gComm_);
  perf_mon_.EventEnd( PromContext::prom_logs_[LOG_METIS] );
  if(verb>0)PetscPrintf(gComm_,"[%d]%s METIS DONE\n",mype,__FUNCT__);
  
  // free buffers
  PetscFree(xadj); 
  PetscFree(vwgts); assert(adjncy[numadjac] == -1);
  PetscFree(adjncy);
  PetscFree(vtxdist);
  if( ewgts != NULL) PetscFree(ewgts);

  ierr = ReduceCounts( lid_part, part_c, newproc_gnode ); CHKERRQ(ierr);  

  return 0;
}


#undef __FUNCT__  
#define __FUNCT__ "ReduceCounts"
/* PromGrid::ReduceCounts ******************************
 * 
 *  - 
 *
 *  INPUT:
 *    - lid_part: new partition1
 *    - part_c: counts for eacj part (out)
 *    - newproc_gnode: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */ 
int PromGrid::ReduceCounts(int lid_part[], int *&part_c, int *&newproc_gnode) 
{
  const int np=comm_.np();
  int ierr, ii, nCommprocs;   assert(isActive());
  MPI_Comm_size( gComm_, &nCommprocs ); 
  int rbuf[PROM_BSZ]; 

  assert(part_c == NULL);
  ierr = PetscMalloc( nCommprocs*sizeof(int),&part_c ); CHKERRQ(ierr);
  PetscMemzero( part_c, nCommprocs * sizeof(int) );
  if(lid_part != NULL){for(ii=0;ii<nLocalNd_;ii++)part_c[lid_part[ii]]++;}
  ierr = PetscMalloc((np+1)*sizeof(int),&newproc_gnode); CHKERRQ(ierr);
  MPI_Allreduce(part_c, newproc_gnode+1, nCommprocs, MPI_INT, MPI_SUM, gComm_);
  // check for zero proc
  for( ii = 0 ; ii < nCommprocs ; ii++ ) {
    if( newproc_gnode[ii+1] == 0 ) { // recover
      PetscPrintf(gComm_,"\t[%d]%s Warning: ndo repartitioning (zero proc)\n",
		  comm_.mype(),__FUNCT__);
      int rank; MPI_Comm_rank( gComm_, &rank ); 
      for(ii=0;ii<nLocalNd_;ii++)lid_part[ii]=rank;
      PetscMemzero( part_c, nCommprocs * sizeof(int) );
      if(lid_part!=NULL){for(ii=0;ii<nLocalNd_;ii++)part_c[lid_part[ii]]++;}
      MPI_Allreduce(part_c,newproc_gnode+1,nCommprocs,MPI_INT,MPI_SUM,gComm_);
      break;
    }
  }
  // new nodes
  for( ii = newproc_gnode[0] = 0 ; ii < nCommprocs ; ii++ ){
    newproc_gnode[ii+1] = newproc_gnode[ii+1] + newproc_gnode[ii];
  }
  //ierr = PetscTrValid(__LINE__, __FUNCT__, __FILE__, __SDIR__);CHKERRQ(ierr);
  if(newproc_gnode[nCommprocs]!=nglobal()){
    PromContext::printArr(stderr,newproc_gnode,nCommprocs,"newproc_gnode[%d]");
    SETERRQ2(PETSC_COMM_SELF,1,"newproc_gnode[np]=%d nglobal()=%d",
	     newproc_gnode[nCommprocs],nglobal());
  }

  if( factor_ != 1 ) { // expand to fill full
    for( ii = 0 ; ii < nCommprocs+1 ; ii++ ) rbuf[ii] = newproc_gnode[ii];
    for( int pp = 1, tt = 1 ; pp < np ; tt++ ){
      for( ii = 0 ; ii < factor_ ; ii++, pp++ ) newproc_gnode[pp] = rbuf[tt];
    }
    assert(newproc_gnode[np] == nglobal());   
  } 
  assert(nCommprocs<=np);

  

  return 0;
}
#endif
